# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""MeSH scheme."""

from galter_subjects_utils.scheme import Scheme


class MeSHScheme(Scheme):
    """MeSH scheme."""

    def __init__(self):
        """Constructor."""
        super().__init__(
            name="MeSH",
            prefix="https://id.nlm.nih.gov/mesh/"
        )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic download functionality."""

from .types_internal import Subject


def converted_to_subjects(rdm_iter, prefix):
    """Adapts an RDMSubject ("Converted") iterable into a Subject iterable.

    :param rdm_iter: Iterable[RDMSubject]

    RDMSubject:

    {
        "id": "<prefix>...",
        "scheme": "...",
        "subject": "..."
    }

    :return: Iterable[Subject]

    Subject:

        id (minus the <prefix> part)
        label
    """
    len_prefix = len(prefix)
    return (
        Subject(
            id=e["id"][len_prefix:] if e["id"].startswith(prefix) else e["id"],
            label=e["subject"]
        ) for e in rdm_iter
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# invenio-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from galter_subjects_utils.deltor import DeltasGenerator
from galter_subjects_utils.scheme import Scheme
from galter_subjects_utils.types_internal import Subject


def subject_A():
    """Subject without qualifier."""
    return Subject(id="subject_a", label="Subject A")


def subject_B():
    """Subject with qualifier."""
    return Subject(id="subject_b", label="Subject B")


class TestScheme(Scheme):
    """Test subjet kind."""

    def __init__(self):
        """Constructor"""
        super().__init__(
            name="SCHEME",
            prefix="https://id.example.org/scheme/",
        )


scheme_for_tests = TestScheme()


# Test Add


def test_deltor_add():
    """Test only addition scenarios."""
    src = [
        subject_A()
    ]
    dst = [
        subject_A(),
        subject_B(),
    ]
    replacement = {}

    deltor = DeltasGenerator(src, dst, scheme_for_tests, replacement)

    add_ops = deltor.generate()

    expected = [
        {
            "id": "https://id.example.org/scheme/subject_b",
            "type": "add",
            "scheme": "SCHEME",
            "subject": "Subject B"
        },
    ]
    assert expected == add_ops


# Test Rename


def subjects_renamed():
    """Subject and its renamed version."""
    return (
        Subject(id="subject_rename", label="Subject Before Rename"),
        Subject(id="subject_rename", label="Subject After Rename"),
    )


def test_deltor_rename():
    """Test rename scenarios only."""
    subject_original, subject_renamed = subjects_renamed()
    src = [
        subject_original,
        subject_A(),
    ]
    dst = [
        subject_A(),
        subject_renamed,
    ]
    replacement = {}
    deltor = DeltasGenerator(src, dst, scheme_for_tests, replacement)

    rename_ops = deltor.generate()

    expected = [
        {
            "id": "https://id.example.org/scheme/subject_rename",
            "type": "rename",
            "scheme": "SCHEME",
            "subject": "Subject Before Rename",
            "new_subject": "Subject After Rename"
        },
    ]
    assert expected == rename_ops


# Test remove


def test_deltor_remove():
    """Test remove scenarios only."""
    subject = subject_B()
    src = [subject]
    dst = []
    replacements = {}
    deltor = DeltasGenerator(src, dst, scheme_for_tests, replacements)

    ops = deltor.generate()

    expected = [
        {
            "id": "https://id.example.org/scheme/" + subject.id,
            "type": "remove",
            "scheme": "SCHEME",
            "subject": "Subject B",
        },
    ]
    assert expected == ops


# Test replace

def subjects_replaced_by_already_present():
    """Subject to be replaced and its replacer for the pre-existing scenario.

    This function doesn't enforce an already present replacer, but
    rather expects it to be used this way.
    The replacer should be passed as part of the source/original subjects.
    """
    return (
        Subject(
            id="D018290Q000145",
            label="Cervical Intraepithelial Neoplasia/classification"
        ),
        Subject(
            id="D002578Q000145",
            label="Uterine Cervical Dysplasia/classification"
        )
    )


def subjects_replaced_by_added():
    """Return a subject and its replacer for the added scenario.

    This function doesn't enforce a newly added replacer, but
    rather expects it to be used this way. The replacer subject should not
    be passed as part of the original subjects.

    Subjects are made up because didn't find actual cases in last 3 years.
    """
    return (
        Subject(
            id="subject-to-replace-by-added",
            label="Subject To Replace By Added"
        ),
        Subject(id="subject-replacer-added", label="Subject Replacer Added")
    )


def subjects_replaced_by_renamed():
    """Return a subject and its replacer for the relabelled scenario.

    This is to be used for the case where a subject is replaced by
    a pre-existing but relabelled subject.
    """
    subject_original, subject_renamed = subjects_renamed()
    return (
        Subject(
            id="subject-to-replace-by-renamed",
            label="Subject To Replace By Renamed"
        ),
        subject_renamed
    )


def test_deltor_replace():
    """Test replace scenarios only (sort-of)."""
    # Case 1: subject is replaced by already existing other subject
    subject_to_be_replaced_by_already_present, subject_replacer_already_present = (  # noqa
        subjects_replaced_by_already_present()
    )
    # Case 2: subject is replaced by newly added other subject
    subject_to_be_replaced_by_added, subject_replacer_added = (
        subjects_replaced_by_added()
    )
    # Case 3: subject is replaced by renamed other subject
    subject_original, subject_renamed = subjects_renamed()
    subject_to_be_replaced_by_renamed, subject_replacer_renamed = (
        subjects_replaced_by_renamed()
    )
    src = [
        subject_to_be_replaced_by_already_present,
        subject_replacer_already_present,
        subject_to_be_replaced_by_added,
        subject_to_be_replaced_by_renamed,
        subject_original,
    ]
    dst = [
        subject_replacer_already_present,
        subject_replacer_added,
        subject_replacer_renamed,
        subject_renamed,
    ]
    replacements = {
        subject_to_be_replaced_by_already_present.label: (
            subject_replacer_already_present.label
        ),
        subject_to_be_replaced_by_added.label: subject_replacer_added.label,
        subject_to_be_replaced_by_renamed.label: subject_replacer_renamed.label,  # noqa
    }
    deltor = DeltasGenerator(src, dst, scheme_for_tests, replacements)

    ops = deltor.generate()

    expected = [
        # contains add and rename operations to account for replacement
        # scenarios
        {
            "id": "https://id.example.org/scheme/subject-replacer-added",
            "type": "add",
            "scheme": "SCHEME",
            "subject": "Subject Replacer Added"
        },
        {
            "id": "https://id.example.org/scheme/subject_rename",
            "type": "rename",
            "scheme": "SCHEME",
            "subject": "Subject Before Rename",
            "new_subject": "Subject After Rename"
        },
        {
            # subject_to_be_replaced_by_already_present.id
            "id": "https://id.example.org/scheme/D018290Q000145",
            "type": "replace",
            "scheme": "SCHEME",
            # subject_to_be_replaced_by_already_present.label
            "subject": "Cervical Intraepithelial Neoplasia/classification",
            # subject_replacer_already_present.id
            "new_id": "https://id.example.org/scheme/D002578Q000145"
        },
        {
            "id": "https://id.example.org/scheme/subject-to-replace-by-added",
            "type": "replace",
            "scheme": "SCHEME",
            "subject": "Subject To Replace By Added",
            "new_id": "https://id.example.org/scheme/subject-replacer-added"
        },
        {
            "id": "https://id.example.org/scheme/subject-to-replace-by-renamed",  # noqa
            "type": "replace",
            "scheme": "SCHEME",
            "subject": "Subject To Replace By Renamed",
            "new_id": "https://id.example.org/scheme/subject_rename"
        },
    ]
    assert expected == ops



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Adapters to convert between different outputs and inputs."""


def generate_replacements(replace_iter):
    """Converts a replace iterable into a replacement dict.

    :param replace_iter: Iterable[ReplaceEntry]

    ReplaceEntry:

    {
        "MH OLD": "...",
        "MH NEW": "...",
        "delete": "...",
        "status": "...",
    }
    """
    return {
        e["MH OLD"]: e["MH NEW"] for e in replace_iter if e["delete"]
    }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2023-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""LCSH Command line tool."""

from datetime import datetime
from functools import wraps
from pathlib import Path

import click
from flask.cli import with_appcontext

from galter_subjects_utils.adapter import converted_to_subjects
from galter_subjects_utils.deltor import DeltasGenerator
from galter_subjects_utils.keeptrace import KeepTrace
from galter_subjects_utils.reader import get_rdm_subjects, read_csv, read_jsonl
from galter_subjects_utils.writer import write_csv

from .adapter import generate_replacements
from .converter import LCSHRDMConverter, raw_to_deprecated
from .downloader import LCSHDownloader
from .scheme import LCSHScheme

defaults = {
    "downloads-dir": Path.cwd(),
    "output-file": Path.cwd(),
}


@click.group()
def lcsh():
    """LCSH related commands."""


def lcsh_download_options(f):
    """Encapsulate common LCSH download options."""

    @click.option(
        "--downloads-dir", "-d",
        type=click.Path(path_type=Path),
        default=defaults["downloads-dir"]
    )
    @click.option(
        "--no-cache",
        default=False,
        help="Re-download even if already downloaded.",
        is_flag=True
    )
    @wraps(f)
    def _wrapped(*args, **kwargs):
        """Wrap f with common download options."""
        return f(*args, **kwargs)

    return _wrapped


def to_lcsh_downloader_kwargs(parameters):
    """To LCSHDownloader kwargs."""
    result = {
        "cache": not parameters["no_cache"],
        "directory": Path.cwd() / parameters["downloads_dir"]
    }
    return result


@lcsh.command("download")
@lcsh_download_options
def lcsh_download(**parameters):
    """Download LCSH files."""
    downloader_kwargs = to_lcsh_downloader_kwargs(parameters)
    downloader = LCSHDownloader(**downloader_kwargs)
    downloader.download()

    print(f"Raw LCSH files written in {parameters['downloads_dir']}/")


def to_lcsh_converter_kwargs(downloader):
    """Provide LCSH converter args from `downloader`."""
    return {
        "topics": read_jsonl(downloader.terms_filepath)
    }


@lcsh.command("file")
@lcsh_download_options
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "subjects_lcsh.csv",
)
def lcsh_file(**parameters):
    """Generate new LCSH subjects file."""
    # Download
    downloader_kwargs = to_lcsh_downloader_kwargs(parameters)
    downloader = LCSHDownloader(**downloader_kwargs)
    downloader.download()

    # Convert
    converter_kwargs = to_lcsh_converter_kwargs(downloader)
    converter = LCSHRDMConverter(**converter_kwargs)
    converted = converter.convert(),
    # Exclude special automated geographic terms
    converted_filtered = (s for s in converted if not s["id"].endswith("-781"))

    # Write
    header = ["id", "scheme", "subject"]
    filepath = write_csv(
        converted_filtered,
        parameters["output_file"],
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"LCSH terms written here {filepath}")


@lcsh.command("deprecated")
@click.argument(
    "downloads-dir",
    type=click.Path(path_type=Path, exists=True, file_okay=False),
)
@click.option("--since", "-s", default=None, help="Filter for YYYY-MM-DD and later.")  # noqa
def lcsh_deprecated(**parameters):
    """Generate CSV file of raw deprecated LCSH topics.

    This file is then parsed by a metadata expert and potentially edited
    for future use in lcsh_deltas.
    """
    downloads_dir = parameters["downloads_dir"].expanduser()
    since = parameters["since"]
    since = datetime.strptime(since, "%Y-%m-%d") if since else None

    fp_of_subjects = downloads_dir / f"subjects.skosrdf.jsonld"
    topics_raw = read_jsonl(fp_of_subjects)
    deprecations = raw_to_deprecated(topics_raw, since)

    header = ["id", "time", "subject", "new_id", "new_subject", "notes"]
    fp_of_deprecated = downloads_dir / "deprecated_lcsh.csv"
    write_csv(
        deprecations,
        fp_of_deprecated,
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"LCSH deprecated written here {fp_of_deprecated}")


@lcsh.command("replacements")
@click.argument(
    "deprecated-file",
    type=click.Path(path_type=Path, exists=True, dir_okay=False),
)
def lcsh_replacements(**parameters):
    """Generate CSV file of replaced LCSH topics.

    Warning: this file has to be reviewed/edited by metadata expert to fill
             out potentially absent "new_subject" field.

    This file is used in lcsh_deltas.
    """
    fp_of_deprecated = parameters["deprecated_file"].expanduser()

    deprecations = read_csv(fp_of_deprecated)
    replacements = (d for d in deprecations if d.get("new_id"))

    header = ["id", "time", "subject", "new_id", "new_subject", "notes"]
    fp_of_replacements = fp_of_deprecated.parent / "replacements_lcsh.csv"
    write_csv(
        replacements,
        fp_of_replacements,
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"LCSH replacements written here {fp_of_replacements}")


@lcsh.command("deltas")
@click.option(
    "--downloads-dir", "-d",
    type=click.Path(path_type=Path),
    default=defaults["downloads-dir"]
)
@click.option(
    "--output-file", "-o",
    type=click.Path(path_type=Path),
    default=defaults["output-file"] / "deltas.csv",
)
@with_appcontext
def lcsh_deltas(**parameters):
    """Write LCSH subject delta operations to file."""
    print("Generating deltas...")
    downloads_dir = parameters["downloads_dir"].expanduser()
    lcsh = LCSHScheme()

    # Source subjects
    subjects_rdm_preexisting = get_rdm_subjects(scheme=lcsh.name)
    src = converted_to_subjects(
        subjects_rdm_preexisting,
        prefix=lcsh.prefix,
    )

    # Destination subjects
    fp_of_subjects = downloads_dir / "subjects.skosrdf.jsonld"
    topics_raw = read_jsonl(fp_of_subjects)
    converted = LCSHRDMConverter(topics_raw).convert()
    # Exclude special automated geographic terms
    converted_filtered = (s for s in converted if not s["id"].endswith("-781"))
    dst = converted_to_subjects(converted_filtered, prefix=lcsh.prefix)

    # Replacements
    fp_of_replacements = downloads_dir / "replacements_lcsh.csv"
    replacements_lcsh = read_csv(fp_of_replacements)
    replacements = generate_replacements(replacements_lcsh)

    ops = (
        DeltasGenerator(
            src_subjects=src,
            dst_subjects=dst,
            scheme=lcsh,
            replacements=replacements
        )
        .generate()
    )
    # in-place
    KeepTrace.mark(
        ops,
        # only can keep trace of *those* anyway
        yes_logic=lambda op: op["type"] in ["rename", "replace", "remove"]
    )

    fp_of_deltas = parameters["output_file"]
    header = [
        "id",
        "type",
        "scheme",
        "subject",
        "new_id",
        "new_subject",
        "keep_trace"
    ]
    write_csv(
        ops,
        fp_of_deltas,
        writer_kwargs={
            "fieldnames": header
        }
    )

    print(f"LCSH deltas written here {fp_of_deltas}")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Northwestern University.
#
# invenio-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from datetime import datetime
from pathlib import Path

from galter_subjects_utils.contrib.lcsh.converter import LCSHRDMConverter, \
    raw_to_deprecated
from galter_subjects_utils.reader import read_jsonl


def test_converter():
    # File is setup to test
    # - regular entries
    # - deprecated entry that is ignored
    # - entry with multiple labels (take the last)
    filepath = Path(__file__).parent / "data" / "fake_lcsh.skosrdf.jsonld"
    topics = list(read_jsonl(filepath))
    converter = LCSHRDMConverter(topics)

    dicts_of_lcsh_terms = [d for d in converter.convert()]

    expected = [
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000011',
            "scheme": "LCSH",
            "subject": "ActionScript (Computer program language)"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh00000014',
            "scheme": "LCSH",
            "subject": "Tacos"
        },
        {
            "id": 'https://id.loc.gov/authorities/subjects/sh90000997',
            "scheme": "LCSH",
            "subject": "Rooms"
        }
    ]
    assert expected == dicts_of_lcsh_terms


def test_deprecated():
    filepath = Path(__file__).parent / "data" / "fake_lcsh.skosrdf.jsonld"
    topics = list(read_jsonl(filepath))

    dicts_of_replacements = [d for d in raw_to_deprecated(topics)]

    expected = [
        # No clear new_id/new_subject
        {
            "id": "https://id.loc.gov/authorities/subjects/sh00000273",
            "time": "2021-07-20T08:25:20",
            "subject": "Child concentration camp inmates",
            "new_id": "",
            "new_subject": "",
            "notes": "This authority record has been deleted because the heading is covered by the subject headings {Child internment camp inmates} (DLC)sh2021004026 and {Child Nazi concentration camp inmates} (DLC)sh2021004027"  # noqa
        },
        # Clear new_id/new_subject
        {
            "id": "https://id.loc.gov/authorities/subjects/sh2008007279",
            "time": "2023-01-23T12:01:15",
            "subject": "Computer games industry",
            "new_id": "https://id.loc.gov/authorities/subjects/sh2006005259",
            "new_subject": "Video games industry",
            "notes": "This authority record has been deleted because the heading is covered by the subject heading {Video games industry} (DLC)sh2006005259"  # noqa
        },
        # Variation: strange formatting to account for
        {
            "id": "https://id.loc.gov/authorities/subjects/sh2006004185",
            "time": "2023-01-23T12:01:15",
            "subject": "Computer games--Law and legislation",
            "new_id": "https://id.loc.gov/authorities/subjects/sh85143203",
            "new_subject": "Video games--Law and legislation",
            "notes": "This authority record has been deleted because the heading is covered by the subject heading {Video games--Law and legislation} (DLC)sh 85143203"  # noqa
        },
        # Variation: different wording + no replacement subject
        {
            "id": "https://id.loc.gov/authorities/subjects/sh2022001344",
            "time": "2023-09-29T18:42:32",
            "subject": "Evangelisch-Lutherse Kerk (Netherlands)--Relations--Nederlandse Hervormde Kerk",  # noqa
            "new_id": "https://id.loc.gov/authorities/subjects/sh2022001319",
            "new_subject": "",
            "notes": "This authority record has been deleted because the heading is covered by an identical subject heading (DLC)sh2022001319"  # noqa
        },
    ]
    assert expected == dicts_of_replacements


def test_deprecated_since():
    filepath = Path(__file__).parent / "data" / "fake_lcsh.skosrdf.jsonld"
    topics = list(read_jsonl(filepath))

    dicts_of_replacements = [
        d for d in raw_to_deprecated(topics, datetime(2023, 9, 29))
    ]

    expected = [
        {
            "id": "https://id.loc.gov/authorities/subjects/sh2022001344",
            "time": "2023-09-29T18:42:32",
            "subject": "Evangelisch-Lutherse Kerk (Netherlands)--Relations--Nederlandse Hervormde Kerk",  # noqa
            "new_id": "https://id.loc.gov/authorities/subjects/sh2022001319",
            "new_subject": "",
            "notes": "This authority record has been deleted because the heading is covered by an identical subject heading (DLC)sh2022001319"  # noqa
        },
    ]
    assert expected == dicts_of_replacements



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022-2024 Northwestern University.
#
# invenio-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from collections import namedtuple
from contextlib import contextmanager
from pathlib import Path
from unittest import mock

from galter_subjects_utils.contrib.lcsh.downloader import LCSHDownloader

# Helpers


@contextmanager
def fake_request_context(url, stream):
    fp = Path(__file__).parent / "data/fake_lcsh.skosrdf.jsonld.gz"
    FakeRequestContext = namedtuple("FakeRequestContext", ["raw"])
    with open(fp, "rb") as f:
        yield FakeRequestContext(raw=f)


# Tests


@mock.patch('galter_subjects_utils.downloader.requests.get')
def test_downloader(patched_get, tmp_path):
    # patch requests.get to return files
    patched_get.side_effect = fake_request_context
    downloads_dir = tmp_path / "downloads"
    downloads_dir.mkdir()

    # No cache by default
    downloader = LCSHDownloader(directory=downloads_dir)
    downloader.download()

    patched_get.assert_called()
    assert (
        downloads_dir / "subjects.skosrdf.jsonld.gz" ==
        downloader.downloaded_filepath
    )
    assert (
        downloads_dir / "subjects.skosrdf.jsonld" ==
        downloader.terms_filepath
    )

    # With cache
    patched_get.reset_mock()
    downloader = LCSHDownloader(directory=downloads_dir, cache=True)
    downloader.download()

    patched_get.assert_not_called()
    assert (
        downloads_dir / "subjects.skosrdf.jsonld.gz" ==
        downloader.downloaded_filepath
    )
    assert (
        downloads_dir / "subjects.skosrdf.jsonld" ==
        downloader.terms_filepath
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Subject's Scheme."""


class Scheme:
    """Parent class to specific subject kinds.

    Meant as a way to get the template method generate_id, and scheme + prefix
    interface.
    """

    def __init__(self, name=None, prefix=None):
        """Constructor."""
        self.name = name
        self.prefix = prefix

    def generate_id(self, identifier):
        """Generate full URL id as expected by InvenioRDM."""
        return f"{self.prefix}{identifier}"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic download functionality."""

from dataclasses import dataclass


@dataclass
class Subject:
    """Minimalist subject data."""

    id: str
    label: str



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Convert MeSH data to InvenioRDM/ops."""

from .scheme import MeSHScheme


class MeSHRDMConverter:
    """Convert MeSH term into RDM subject dict."""

    def __init__(self, topics, qualifiers_mapping=None):
        """Constructor.

        :param topics: MeSH Topics iterable
        :type topics: iterable[dict]
        :param qualifiers: MeSH qualifiers iterable, defaults to None
        :type qualifiers: iterable[dict], optional
        """
        self.topics = topics
        self.qualifiers_mapping = qualifiers_mapping
        self.mesh = MeSHScheme()

    def __iter__(self):
        """Iterate over converted entries.

        DEPRECATED.
        """
        yield from self.convert()

    def convert(self):
        """Iterator over converted entries."""
        for topic in self.topics:
            yield {
                "id": self.mesh.generate_id(topic['UI']),
                "scheme": self.mesh.name,
                "subject": topic['MH']
            }

            if not self.qualifiers_mapping:
                continue

            for qid in topic.get("AQ", []):
                qualifier = self.qualifiers_mapping.get(qid)

                if not qualifier:
                    continue

                yield {
                    "id": self.mesh.generate_id(topic['UI'] + qualifier['UI']),
                    "scheme": self.mesh.name,
                    "subject": topic['MH'] + "/" + qualifier['SH']
                }



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic conversion functionality."""

import re
from datetime import datetime

from .scheme import LCSHScheme


class LCSHRDMConverter:
    """Convert LCSH term into RDM subjects dict."""

    def __init__(self, topics):
        """Constructor.

        :param topics: LCSH Topics iterable
        :type topics: iterable[dict]
        """
        self.topics = topics
        self.lcsh = LCSHScheme()

    def __iter__(self):
        """Iterate over terms.

        DEPRECATED.
        """
        yield from self.convert()

    def convert(self):
        """Iterator over converted entries."""
        for topic in self.topics:
            entry = self.extract_entry(topic)
            if entry:
                yield entry

    def extract_entry(self, topic):
        """Extract relevant dict from LCSH skos json-ld dict."""
        # Find id and label
        id_suffix = topic["@id"]
        info_entry = next(
            (
                e for e in topic["@graph"]
                if (
                    e.get("@id").endswith(id_suffix) and
                    e.get("@type") == "skos:Concept"
                )
            ),
            None
        )

        if not info_entry:
            return {}

        entry = {
            "id": info_entry["@id"].replace("http://", "https://"),
            "scheme": self.lcsh.name
        }
        pref_label = info_entry["skos:prefLabel"]
        if isinstance(pref_label, dict):
            subject = pref_label["@value"]
        elif isinstance(pref_label, list):
            subject = pref_label[-1]["@value"]
        else:
            raise Exception("invalid skos:prefLabel")
        entry["subject"] = subject

        return entry


def find_main_node(topic):
    """Find main topic node among topic's skos graph.

    :param topic: skos topic
    :type topic: dict
    """
    id_suffix = topic["@id"]
    node_main = next(
        (
            n for n in topic["@graph"]
            if n.get("@id", "").endswith(id_suffix)
        ),
        None
    )
    if not node_main:
        raise Exception("No main node for {id_suffix}.")
    return node_main


def find_deprecation_node(topic, node_main):
    """Find deprecation node among topic's skos graph.

    :param topic: skos topic
    :type topic: dict
    :param node_main: main skos graph node
    :type node_main: dict
    """
    for node in topic["@graph"]:
        is_deprecated = node.get("cs:changeReason") == "deprecated"
        subject_of_change = node.get("cs:subjectOfChange", {}).get("@id")
        # just to be sure
        if is_deprecated and node_main.get("@id") == subject_of_change:
            return node
    return {}


def is_deprecated(topic):
    """Filter for deprecated topics.

    :param topic: skos topic
    :type topic: dict
    """
    # Find main node
    node_main = find_main_node(topic)

    # Find deprecation
    return bool(find_deprecation_node(topic, node_main))


def happened_since(topic, since=None):
    """Filter for topic having happened after since.

    :param topic: skos topic
    :type topic: dict
    :param since: date since to filter by (included)
    :type since: datetime.datetime
    """
    if not since:
        return True
    node_main = find_main_node(topic)
    node_deprecation = find_deprecation_node(topic, node_main)
    date_created_str = node_deprecation.get("cs:createdDate", {}).get("@value")
    date_created = datetime.strptime(date_created_str, "%Y-%m-%dT%H:%M:%S")
    return since <= date_created


def find_explanation(nodes_of_change_note):
    """Find explaination among skos:changeNote nodes."""
    for n in nodes_of_change_note:
        explanation = n.get("@value")
        if explanation:
            return explanation
    return ""


pattern = r"covered by .+ subject heading( {(?P<subject_new>.+)})? .+(?P<id_new>sh\s?\d+)$"  # noqa
regex_for_replacement = re.compile(pattern)


def extract_replacement_id_subject(notes):
    """Returns (id_new, subject_new) tuple from notes or ("", "") else."""
    matches = regex_for_replacement.search(notes)
    if not matches:
        return "", ""
    id_new = matches.group("id_new") or ""
    id_new = id_new.replace(" ", "")  # get rid of potential space after 'sh'
    lcsh = LCSHScheme()
    id_new = lcsh.generate_id(id_new) if id_new else id_new
    subject_new = matches.group("subject_new") or ""
    return id_new, subject_new


def to_raw_replacement(topic):
    """Fill out replacement dict.

    :param topic: skos topic
    :type topic: dict
    """
    node_main = find_main_node(topic)
    node_deprecation = find_deprecation_node(topic, node_main)
    notes = find_explanation(node_main.get("skos:changeNote"))
    id_new, subject_new = extract_replacement_id_subject(notes)
    return {
        "id": node_main.get("@id").replace("http://", "https://"),
        "time": node_deprecation.get("cs:createdDate", {}).get("@value"),
        "subject": node_main.get("skosxl:literalForm", {}).get("@value"),
        "new_id": id_new,
        "new_subject": subject_new,
        "notes": notes,
    }


def raw_to_deprecated(topics, since=None):
    """Iterator over deprecated topics and their relevant information.

    :param topics: LCSH Topics iterable
    :type topics: iterable[dict]
    :param since: filter for this date (included) and later
    :type since: datetime.datetime

    :yields: Replacement dicts of the shape:
        {
            "id": ...,
            "time": ...,
            "subject": ...,
            "new_id": ...,
            "new_subject": ...,
            "notes": ...
        }
    """
    for topic in topics:
        # filter for deprecated
        if not is_deprecated(topic):
            continue

        # filter for date
        if not (happened_since(topic, since)):
            continue

        # fill out result
        yield to_raw_replacement(topic)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Test MeSH-related conversion functionality."""

from galter_subjects_utils.contrib.mesh.converter import MeSHRDMConverter


def test_converter():
    topics = [{
        'MH': 'Seed Bank',
        'DC': '1',
        'AQ': ['CL', 'EC'],
        'UI': 'D000068098'
    }]
    qualifiers_mapping = {
        "CL": {
            "QA": "CL",
            "SH": "classification",
            "UI": "Q000145"
        },
        "EC": {
            "QA": "EC",
            "SH": "economics",
            "UI": "Q000191"
        }
    }

    converter = MeSHRDMConverter(topics, qualifiers_mapping)
    objects = [o for o in converter]

    expected = [
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098',
            "scheme": "MeSH",
            "subject": "Seed Bank"
        },
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098Q000145',
            "scheme": "MeSH",
            "subject": "Seed Bank/classification"
        },
        {
            "id": 'https://id.nlm.nih.gov/mesh/D000068098Q000191',
            "scheme": "MeSH",
            "subject": "Seed Bank/economics"
        },
    ]
    assert expected == objects



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2023 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Generic download functionality."""

import gzip
import shutil

import requests


def download_file(url, dest):
    """Download a file."""
    with requests.get(url, stream=True) as req:
        with open(dest, 'wb') as f:
            shutil.copyfileobj(req.raw, f)

    return dest


def extract_file(in_filepath, out_filepath):
    """Extract gzipped file."""
    with gzip.open(in_filepath, 'rb') as f_in:
        with open(out_filepath, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)

    return out_filepath



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021-2024 Northwestern University.
#
# galter-subjects-utils is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Download LCSH files."""

from galter_subjects_utils.downloader import download_file, extract_file


class LCSHDownloader:
    """Download LCSH file."""

    def __init__(self, directory, cache=False):
        """Constructor."""
        self.base_url = "https://id.loc.gov/download/authorities/subjects.skosrdf.jsonld.gz"  # noqa
        self.directory = directory
        self.cache = cache

    @property
    def downloaded_filepath(self):
        """Filepath of downloaded gzipped file."""
        return self.directory / self.base_url.rsplit("/")[-1]

    @property
    def terms_filepath(self):
        """Filepath of unzipped terms file."""
        return self.downloaded_filepath.with_suffix("")

    def download(self):
        """Download LCSH files of interest."""
        if not self.cache or not self.downloaded_filepath.exists():
            download_file(self.base_url, self.downloaded_filepath)
            # sneak extraction in caching: bit of a cheat, but ok for now
            extract_file(
                self.downloaded_filepath, self.terms_filepath
            )
